import fpuchain.fpucuda as fpucuda
import fpuchain.fpucpu as fpucpu
import numpy
import time
import math

class FpuBase:
    #new FPU solver for a chain of nvar particles
    def __init__(self,nvar):
        pass
    #close solver
    def close(self):
        pass
    #initial condition in normal mode m
    def init(self,m):
        pass
    #run solver
    #   h : time step
    #   n1 : number of steps in a block
    #   n2 : number of blocks = number of stored values
    #   b : amplitude of non-linear term
    #   n : power of non-linear term
    #   nm : number of modes
    def run(self,h,n1,n2,b,n,nm):
        pass
    
    
#FPU chain with fixed boundaries, CUDA implementation
class FpuCuda(FpuBase):
    def __init__(self,nvar,dev=0):
        self.nvar = nvar
        fpucuda.fpu_open(nvar,dev)
    def close(self):
        fpucuda.fpu_close()
    def init(self,m):
        q0=numpy.empty(shape=self.nvar,dtype=numpy.float32)
        p0=numpy.empty(shape=self.nvar,dtype=numpy.float32)
        for k in range(self.nvar):
            q0[k] = 1.0*math.sin((k+1)*m*math.pi/(self.nvar+1))
            p0[k] = 0.0
        fpucuda.fpu_init(q0,p0)
    def run(self,h,n1,n2,b,n,nm):
        if nm > self.nvar:
            nm = self.nvar
        self.t = numpy.arange(n2)*h*n1
        fpucuda.fpu_set_cst(b,n,h)
        self.qmat=numpy.empty(shape=(n2,self.nvar),dtype=numpy.float32)
        self.pmat=numpy.empty(shape=(n2,self.nvar),dtype=numpy.float32)
        self.emat=numpy.empty(shape=(n2,1),dtype=numpy.float32)
        self.emodemat=numpy.empty(shape=(n2,self.nvar),dtype=numpy.float32)
        ta = time.time()
        for i2 in range(n2):
            fpucuda.fpu_steps(n1,self.qmat[i2],self.pmat[i2],self.emat[i2],self.emodemat[i2],nm)
            print "%d sur %d"%(i2,n2)
        self.T = time.time()-ta
        print "duree du calcul (s) : %f"%(self.T)
        self.deltaE = numpy.std(self.emat)/numpy.average(self.emat)
        print "Ecart type relatif de E : %e"%self.deltaE

#FPU chain width boundary periodic conditions, CUDA implementation
class FpuPerCuda(FpuBase):
    def __init__(self,nvar,dev=0):
        self.nvar = nvar
        fpucuda.fpuper_open(nvar,dev)
    def close(self):
        fpucuda.fpuper_close()
    def init(self,m):
        q0=numpy.empty(shape=self.nvar,dtype=numpy.float32)
        p0=numpy.empty(shape=self.nvar,dtype=numpy.float32)
        wm=2*math.sin(math.pi*m/self.nvar)
        for k in range(self.nvar):
            q0[k] = math.cos(2*math.pi*(k+1)*m/self.nvar)
            p0[k] = wm*math.sin(2*math.pi*(k+1)*m/self.nvar)
        fpucuda.fpuper_init(q0,p0)
        self.cel=self.nvar/(math.pi*m)*math.sin(math.pi*m/self.nvar)
    def run(self,h,n1,n2,b,n,nm):
        if nm > self.nvar:
            nm = self.nvar
        self.t = numpy.arange(n2)*h*n1
        fpucuda.fpu_set_cst(b,n,h)
        self.qmat=numpy.empty(shape=(n2,self.nvar),dtype=numpy.float32)
        self.pmat=numpy.empty(shape=(n2,self.nvar),dtype=numpy.float32)
        self.emat=numpy.empty(shape=(n2,1),dtype=numpy.float32)
        self.emodemat=numpy.empty(shape=(n2,self.nvar),dtype=numpy.float32)
        self.qstat=numpy.empty(shape=(n2,self.nvar),dtype=numpy.float32)
        self.pstat=numpy.empty(shape=(n2,self.nvar),dtype=numpy.float32)
        ta = time.time()
        for i2 in range(n2):
            fpucuda.fpuper_steps(n1,self.qmat[i2],self.pmat[i2],self.emat[i2],self.emodemat[i2],nm)
            dec = numpy.int(self.cel*h*n1*(i2+1))
            self.qstat[i2] = numpy.roll(self.qmat[i2],-dec)
            self.pstat[i2] = numpy.roll(self.pmat[i2],-dec)
            print "%d sur %d"%(i2,n2)
        self.T = time.time()-ta
        print "duree du calcul (s) : %f"%(self.T)
        self.deltaE = numpy.std(self.emat)/numpy.average(self.emat)
        print "Ecart type relatif de E : %e"%self.deltaE

#FPU chain with fixed boundaries, CPU implementation
class Fpu(FpuBase):
    def __init__(self,nvar):
        self.nvar = nvar
        fpucpu.fpu_open(nvar)
    def close(self):
        fpucpu.fpu_close()
    def init(self,m):
        q0=numpy.empty(shape=self.nvar,dtype=numpy.float32)
        p0=numpy.empty(shape=self.nvar,dtype=numpy.float32)
        for k in range(self.nvar):
            q0[k] = 1.0*math.sin((k+1)*m*math.pi/(self.nvar+1))
            p0[k] = 0.0
        fpucpu.fpu_init(q0,p0)
    def run(self,h,n1,n2,b,n,nm):
        if nm > self.nvar:
            nm = self.nvar
        self.t = numpy.arange(n2)*h*n1
        fpucpu.fpu_set_cst(b,n,h)
        self.qmat=numpy.empty(shape=(n2,self.nvar),dtype=numpy.float32)
        self.pmat=numpy.empty(shape=(n2,self.nvar),dtype=numpy.float32)
        self.emat=numpy.empty(shape=(n2,1),dtype=numpy.float32)
        self.emodemat=numpy.empty(shape=(n2,self.nvar),dtype=numpy.float32)
        ta = time.time()
        for i2 in range(n2):
            fpucpu.fpu_steps(n1,self.qmat[i2],self.pmat[i2],self.emat[i2],self.emodemat[i2],nm)
            print "%d sur %d"%(i2,n2)
        self.T = time.time()-ta
        print "duree du calcul (s) : %f"%(self.T)
        self.deltaE = numpy.std(self.emat)/numpy.average(self.emat)
        print "Ecart type relatif de E : %e"%self.deltaE

#FPU chain width boundary periodic conditions, CPU implementation
class FpuPer(FpuBase):
    def __init__(self,nvar):
        self.nvar = nvar
        fpucpu.fpuper_open(nvar)
    def close(self):
        fpucpu.fpuper_close()
    def init(self,m):
        q0=numpy.empty(shape=self.nvar,dtype=numpy.float32)
        p0=numpy.empty(shape=self.nvar,dtype=numpy.float32)
        wm=2*math.sin(math.pi*m/self.nvar)
        for k in range(self.nvar):
            q0[k] = math.cos(2*math.pi*(k+1)*m/self.nvar)
            p0[k] = wm*math.sin(2*math.pi*(k+1)*m/self.nvar)
        fpucpu.fpuper_init(q0,p0)
        self.cel=self.nvar/(math.pi*m)*math.sin(math.pi*m/self.nvar)
    def run(self,h,n1,n2,b,n,nm):
        if nm > self.nvar:
            nm = self.nvar
        self.t = numpy.arange(n2)*h*n1
        fpucpu.fpu_set_cst(b,n,h)
        self.qmat=numpy.empty(shape=(n2,self.nvar),dtype=numpy.float32)
        self.pmat=numpy.empty(shape=(n2,self.nvar),dtype=numpy.float32)
        self.emat=numpy.empty(shape=(n2,1),dtype=numpy.float32)
        self.emodemat=numpy.empty(shape=(n2,self.nvar),dtype=numpy.float32)
        self.qstat=numpy.empty(shape=(n2,self.nvar),dtype=numpy.float32)
        self.pstat=numpy.empty(shape=(n2,self.nvar),dtype=numpy.float32)
        ta = time.time()
        for i2 in range(n2):
            fpucpu.fpuper_steps(n1,self.qmat[i2],self.pmat[i2],self.emat[i2],self.emodemat[i2],nm)
            dec = numpy.int(self.cel*h*n1*(i2+1))
            self.qstat[i2] = numpy.roll(self.qmat[i2],-dec)
            self.pstat[i2] = numpy.roll(self.pmat[i2],-dec)
            print "%d sur %d"%(i2,n2)
        self.T = time.time()-ta
        print "duree du calcul (s) : %f"%(self.T)
        self.deltaE = numpy.std(self.emat)/numpy.average(self.emat)
        print "Ecart type relatif de E : %e"%self.deltaE
