/*---------------------------------------------------------------------
Copyright Frédéric Legrand (2011)

This software is a computer program whose purpose is to solve differential
equations of non-linear oscillator chain (Fermi-Pasta-Ulam problem).

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

-------------------------------------------------------------------------*/
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "fpu.h"

#define PI 3.1415926

static int nvar;
static float b,n,h;
static float *q, *p, *dp, en;
static float *A, *omeg2, *emode, *xi, *dxi;
static float *reA, *imA, *rexi, *imxi, *redxi, *imdxi;

/* FPU chain with fixed boundary conditions */

static void dPotential(void) {
    float x1,x2;
    int k;
    x1=q[0]; x2=q[1]-q[0];
    dp[0] = x1-x2+b*(pow(x1,n)-pow(x2,n));
    for (k=1; k<nvar-1; k++) {
        x1 = q[k]-q[k-1]; x2=-q[k]+q[k+1];
        dp[k] = x1-x2+b*(pow(x1,n)-pow(x2,n));
    }
    x1 = q[k]-q[k-1]; x2=-q[k];
    dp[k] = x1-x2+b*(pow(x1,n)-pow(x2,n));
}

static void verletStep(void) {
            int k;
            dPotential();
            for (k=0; k<nvar; k++) p[k] -= 0.5*h*dp[k];
            for (k=0; k<nvar; k++) q[k] += h*p[k];
            dPotential();
            for (k=0; k<nvar; k++) p[k] -= 0.5*h*dp[k];
}

static void energy(void) {
            float x1,x2;
            int k;
            x1=q[0];
            en = 0.5*p[0]*p[0]+0.5*(x1*x1)+1.0/(n+1)*b*pow(x1,n+1);
            for (k=1; k<nvar-1; k++) {
                   x1=q[k]-q[k-1];
                   en += 0.5*p[k]*p[k]+0.5*(x1*x1)+1.0/(n+1)*b*pow(x1,n+1);
            }
            x1=q[k]-q[k-1]; x2=-q[k];
            en += 0.5*p[k]*p[k]+0.5*(x1*x1+x2*x2)+1.0/(n+1)*b*(pow(x1,n+1)+pow(x2,n+1));
}

static void verletMultiSteps(int nsteps) {
    int i;
    for (i=0; i<nsteps; i++) verletStep();
    energy();
}

static void modeEnergy(int nm) {
    int i,j;
    int index;
    for (i=0; i<nm; i++) {
        xi[i] = 0.0f;
        dxi[i] = 0.0f;
        for (j=0; j<nvar;j++) {
            index = j+nvar*i;
            xi[i] += A[index]*q[j];
            dxi[i] += A[index]*p[j];
        }
        emode[i] = 0.5*(dxi[i]*dxi[i]+omeg2[i]*xi[i]*xi[i]);
    }
}

int fpu_cpu_open(int _nvar) {
    float r,s;
    int i,j;
    nvar = _nvar;
    q = malloc(nvar*sizeof(float));
    p = malloc(nvar*sizeof(float));
    dp = malloc(nvar*sizeof(float));
    omeg2 = malloc(nvar*sizeof(float));
    emode = malloc(nvar*sizeof(float));
    xi = malloc(nvar*sizeof(float));
    dxi = malloc(nvar*sizeof(float));
    A = malloc(nvar*nvar*sizeof(float));
    A = (float*)malloc(nvar*nvar*sizeof(float));
    r = PI/(nvar+1);
    s = sqrt(2.0/(nvar+1));
    for (i=0; i<nvar; i++)
        for (j=0; j<nvar; j++) {
            A[i+nvar*j] = s*sin(r*(i+1)*(j+1));
    }
    r=PI/(2*(nvar+1));
    for (i=0; i<nvar; i++) omeg2[i] = 4*pow(sin((i+1)*r),2);
    return 0;
}

int fpu_cpu_close() {
    free(q); free(p); free(dp); free(omeg2); free(emode); free(xi); free(dxi); free(A);
    return 0;
}

int fpu_cpu_init(float *q0, float *p0) {
    memcpy(q,q0,nvar*sizeof(float));
    memcpy(p,p0,nvar*sizeof(float));
    return 0;
}

int fpu_cpu_set_cst(float _b, float _n , float _h) {
    b = _b;
    n = _n;
    h = _h;
    return 0;  
}

int fpu_cpu_steps(int n1, float *q1, float *p1, float *en1, float *emode1, int nm) {
    verletMultiSteps(n1);
    modeEnergy(nm);
    memcpy(q1,q,nvar*sizeof(float));
    memcpy(p1,p,nvar*sizeof(float));
    en1[0] = en;
    memcpy(emode1,emode,nvar*sizeof(float));
    return 0;
}

/*FPU chain with periodic boundary conditions*/

static void dPotential_per(void) {
    float x1,x2;
    int k;
    x1=q[0]-q[nvar-1]; x2=q[1]-q[0];
    dp[0] = x1-x2+b*(pow(x1,n)-pow(x2,n));
    for (k=1; k<nvar-1; k++) {
        x1 = q[k]-q[k-1]; x2=-q[k]+q[k+1];
        dp[k] = x1-x2+b*(pow(x1,n)-pow(x2,n));
    }
    x1 = q[k]-q[k-1]; x2=-q[k]+q[0];
    dp[k] = x1-x2+b*(pow(x1,n)-pow(x2,n));
}

static void verletStep_per(void) {
            int k;
            dPotential_per();
            for (k=0; k<nvar; k++) p[k] -= 0.5*h*dp[k];
            for (k=0; k<nvar; k++) q[k] += h*p[k];
            dPotential_per();
            for (k=0; k<nvar; k++) p[k] -= 0.5*h*dp[k];
}

static void energy_per(void) {
            float x1;
            int k;
            x1=q[0]-q[nvar-1];
            en = 0.5*p[0]*p[0]+0.5*(x1*x1)+1.0/(n+1)*b*pow(x1,n+1);
            for (k=1; k<nvar; k++) {
                   x1=q[k]-q[k-1];
                   en += 0.5*p[k]*p[k]+0.5*(x1*x1)+1.0/(n+1)*b*pow(x1,n+1);
            }
}

static void verletMultiSteps_per(int nsteps) {
    int i;
    for (i=0; i<nsteps; i++) verletStep_per();
    energy_per();
}

static void modeEnergy_per(int nm) {
    int i,j;
    int index;
    for (i=0; i<nm; i++) {
        rexi[i] = 0.0f; imxi[i] = 0.0f;
        redxi[i] = 0.0f; imdxi[i] = 0.0f;
        for (j=0; j<nvar;j++) {
            index = j+nvar*i;
            rexi[i] += reA[index]*q[j];
            imxi[i] += imA[index]*q[j];
            redxi[i] += reA[index]*p[j];
            imdxi[i] += imA[index]*p[j];
        }
        emode[i] = 0.5*(redxi[i]*redxi[i]+imdxi[i]*imdxi[i]+omeg2[i]*(rexi[i]*rexi[i]+imxi[i]*imxi[i]));
    }
}

int fpuper_cpu_open(int _nvar) {
    float r,s;
    int i,j;
    nvar = _nvar;
    q = malloc(nvar*sizeof(float));
    p = malloc(nvar*sizeof(float));
    dp = malloc(nvar*sizeof(float));
    omeg2 = malloc(nvar*sizeof(float));
    emode = malloc(nvar*sizeof(float));
    rexi = malloc(nvar*sizeof(float));
    redxi = malloc(nvar*sizeof(float));
    imxi = malloc(nvar*sizeof(float));
    imdxi = malloc(nvar*sizeof(float));
    reA = malloc(nvar*nvar*sizeof(float));
    imA = malloc(nvar*nvar*sizeof(float));
    r = PI*2/nvar;
    s = sqrt(1.0/nvar);
    for (i=0; i<nvar; i++)
        for (j=0; j<nvar; j++) {
            reA[i+nvar*j] = s*cos(r*(i+1)*(j+1));
            imA[i+nvar*j] = -s*sin(r*(i+1)*(j+1));
    }
    r=PI/nvar;
    for (i=0; i<nvar; i++) omeg2[i] = 4*pow(sin((i+1)*r),2);
    return 0;
}

int fpuper_cpu_close() {
    free(q); free(p); free(dp); free(omeg2); free(emode); free(rexi); free(imxi); free(redxi); free(imdxi); free(reA); free(imA);
    return 0;
}

int fpuper_cpu_init(float *q0, float *p0) {
    memcpy(q,q0,nvar*sizeof(float));
    memcpy(p,p0,nvar*sizeof(float));
    return 0;
}

int fpuper_cpu_set_cst(float _b, float _n , float _h) {
    b = _b;
    n = _n;
    h = _h;
    return 0;  
}

int fpuper_cpu_steps(int n1, float *q1, float *p1, float *en1, float *emode1, int nm) {
    verletMultiSteps_per(n1);
    modeEnergy_per(nm);
    memcpy(q1,q,nvar*sizeof(float));
    memcpy(p1,p,nvar*sizeof(float));
    en1[0] = en;
    memcpy(emode1,emode,nvar*sizeof(float));
    return 0;
}

