/*---------------------------------------------------------------------
Copyright Frédéric Legrand (2011)
  
This software is a computer program whose purpose is to solve differential
equations of non-linear oscillator chain (Fermi-Pasta-Ulam problem).

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,  
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".  

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 
 
The fact that you are presently reading this means that you have had 
knowledge of the CeCILL license and that you accept its terms.
 
-------------------------------------------------------------------------*/
#include <Python.h> 
#include <numpy/npy_common.h>
#include <numpy/ndarrayobject.h>
#include "fpukernel.h"

#define KERNEL_ERROR(e) switch(e) {\
    case FPU_KERNEL_NO_CUDA : PyErr_SetString(PyExc_ValueError,"no CUDA device available"); return NULL;\
    case FPU_KERNEL_NVAR_TOO_LARGE : PyErr_SetString(PyExc_ValueError,"chain size too large for this CUDA device"); return NULL;\
    case FPU_KERNEL_CUDA_ERROR :  PyErr_SetString(PyExc_ValueError,"CUDA error"); return NULL;\
}
#define ARG_ERROR(e) if (e) {PyErr_SetString(PyExc_ValueError,"bad arguments"); return NULL;}

static PyObject* py_fpu_open(PyObject *self, PyObject *args) {
    int nvar,dev;
    ARG_ERROR(!PyArg_ParseTuple(args,"ii",&nvar,&dev))
    KERNEL_ERROR(fpu_open(nvar,dev))
    Py_RETURN_NONE;
}

static PyObject* py_fpuper_open(PyObject *self, PyObject *args) {
    int nvar,dev;
    ARG_ERROR(!PyArg_ParseTuple(args,"ii",&nvar,&dev))
    KERNEL_ERROR(fpuper_open(nvar,dev))  
    Py_RETURN_NONE;
}
 
static PyObject* py_fpu_close(PyObject *self, PyObject *args) {
    KERNEL_ERROR(fpu_close())
    Py_RETURN_NONE;  
}        
   
static PyObject* py_fpuper_close(PyObject *self, PyObject *args) {
    KERNEL_ERROR(fpuper_close())
    Py_RETURN_NONE; 
}
 
static PyObject* py_fpu_set_cst(PyObject *self, PyObject *args) {
    float n,b,h;
    ARG_ERROR(!PyArg_ParseTuple(args,"fff",&b,&n,&h))
    KERNEL_ERROR(fpu_set_cst(b,n,h))
    Py_RETURN_NONE;
}

static PyObject* py_fpuper_set_cst(PyObject *self, PyObject *args) {
    float n,b,h;
    ARG_ERROR(!PyArg_ParseTuple(args,"fff",&b,&n,&h))
    KERNEL_ERROR(fpuper_set_cst(b,n,h))
    Py_RETURN_NONE;
}


static PyObject* py_fpu_init(PyObject *self, PyObject *args) {
    int nq,np;
    PyObject *q0_array, *p0_array;
    float *q0_data, *p0_data;
    npy_intp c[1];
    ARG_ERROR(!PyArg_ParseTuple(args,"OO",&q0_array,&p0_array))
    nq = PyArray_SIZE(q0_array);
    np = PyArray_SIZE(p0_array);
    if (nq!=np) {
        PyErr_SetString(PyExc_ValueError,"incompatible array size");
        return NULL;
    }
    c[0] = 0; 
    q0_data = (float*)PyArray_GetPtr((PyArrayObject*)q0_array,c);
    p0_data = (float*)PyArray_GetPtr((PyArrayObject*)p0_array,c);
    KERNEL_ERROR(fpu_init(q0_data,p0_data))
    Py_RETURN_NONE;
}

static PyObject* py_fpuper_init(PyObject *self, PyObject *args) {
    int nq,np;
    PyObject *q0_array, *p0_array;
    float *q0_data, *p0_data;
    npy_intp c[1];
    ARG_ERROR(!PyArg_ParseTuple(args,"OO",&q0_array,&p0_array))
    nq = PyArray_SIZE(q0_array);
    np = PyArray_SIZE(p0_array);
    if (nq!=np) {
        PyErr_SetString(PyExc_ValueError,"incompatible array size");
        return NULL;
    }
    c[0] = 0; 
    q0_data = (float*)PyArray_GetPtr((PyArrayObject*)q0_array,c);
    p0_data = (float*)PyArray_GetPtr((PyArrayObject*)p0_array,c);
    KERNEL_ERROR(fpuper_init(q0_data,p0_data))
    Py_RETURN_NONE;
} 

static PyObject* py_fpu_steps(PyObject *self, PyObject *args) {
    int n1,nm;
    PyObject *q_array, *p_array, *e_array, *emode_array;
    float *q_data, *p_data, *e_data, *emode_data;
    npy_intp c[1];
    ARG_ERROR(!PyArg_ParseTuple(args,"iOOOOi",&n1,&q_array,&p_array,&e_array,&emode_array,&nm))
    c[0] = 0; 
    q_data = (float*)PyArray_GetPtr((PyArrayObject*)q_array,c);
    p_data = (float*)PyArray_GetPtr((PyArrayObject*)p_array,c);
    e_data = (float*)PyArray_GetPtr((PyArrayObject*)e_array,c);
    emode_data = (float*)PyArray_GetPtr((PyArrayObject*)emode_array,c);
    KERNEL_ERROR(fpu_steps(n1,q_data,p_data,e_data,emode_data,nm))
    Py_RETURN_NONE;
} 

static PyObject* py_fpuper_steps(PyObject *self, PyObject *args) {
    int n1,nm;
    PyObject *q_array, *p_array, *e_array, *emode_array;
    float *q_data, *p_data, *e_data, *emode_data;
    npy_intp c[1];
    ARG_ERROR(!PyArg_ParseTuple(args,"iOOOOi",&n1,&q_array,&p_array,&e_array,&emode_array,&nm))
    c[0] = 0; 
    q_data = (float*)PyArray_GetPtr((PyArrayObject*)q_array,c);
    p_data = (float*)PyArray_GetPtr((PyArrayObject*)p_array,c); 
    e_data = (float*)PyArray_GetPtr((PyArrayObject*)e_array,c);
    emode_data = (float*)PyArray_GetPtr((PyArrayObject*)emode_array,c);
    KERNEL_ERROR(fpuper_steps(n1,q_data,p_data,e_data,emode_data,nm))
    Py_RETURN_NONE; 
}
       

static PyMethodDef fpucudaMethods[] = {     
    {"fpu_open",py_fpu_open,METH_VARARGS},
    {"fpu_close",py_fpu_close,METH_NOARGS},
    {"fpu_init",py_fpu_init,METH_VARARGS},  
    {"fpu_set_cst",py_fpu_set_cst,METH_VARARGS},   
    {"fpu_steps",py_fpu_steps,METH_VARARGS},
    {"fpuper_open",py_fpuper_open,METH_VARARGS},  
    {"fpuper_close",py_fpuper_close,METH_NOARGS},
    {"fpuper_init",py_fpuper_init,METH_VARARGS},
    {"fpuper_set_cst",py_fpuper_set_cst,METH_VARARGS},
    {"fpuper_steps",py_fpuper_steps,METH_VARARGS}, 
    {NULL,NULL}   
    };            
         
void initfpucuda(void) {
    Py_InitModule("fpucuda",fpucudaMethods);
    import_array();         
} 
