/*---------------------------------------------------------------------
Copyright Frédéric Legrand (2011)

This software is a computer program whose purpose is to solve differential
equations of non-linear oscillator chain (Fermi-Pasta-Ulam problem).

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

-------------------------------------------------------------------------*/
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#define FPU_KERNEL_NO_CUDA -1
#define FPU_KERNEL_NVAR_TOO_LARGE -2
#define FPU_KERNEL_CUDA_ERROR -3

#define PI 3.1415926

#define HANDLE_ERROR(err) if (err != cudaSuccess) { \
    printf( "%s in %s at line %d\n", cudaGetErrorString( err ), __FILE__, __LINE__ );\
    return FPU_KERNEL_CUDA_ERROR;\
    }

static int nvar;
static float *dev_A, *dev_reA, *dev_imA, *dev_q, *dev_p, *dev_en, *dev_omeg2, *dev_emode;
static float *dev_dp, *dev_xib, *dev_dxib, *dev_rexib, *dev_imxib, *dev_redxib, *dev_imdxib;
static float b, n, h;
static int cpmode = 0;
static int maxThreads;
static int nblocks, nthreads;

#define CPMODE_LOCAL_MEM 1
#define CPMODE_GLOBAL_MEM 2

    
/* FPU chain with fixed boundary conditions */
__device__ void dPotential(int k, int nvar, float b, float n, float *q, float *dp) {
            float x1,x2;
            if (k==0) {x1=q[0]; x2=q[1]-q[0];}
            else if (k==nvar-1) {x1 = q[k]-q[k-1]; x2=-q[k];}
            else {x1=q[k]-q[k-1]; x2=-q[k]+q[k+1];}
            dp[k] = x1-x2+b*(powf(x1,n)-powf(x2,n));
}

/*integrator loop with local memory*/
__device__ void verletStep(int k, float h, int nvar, float b, float n, float *q, float *p, float *dp) {
            dPotential(k,nvar,b,n,q,dp);
            p[k] -= 0.5*h*dp[k];
            q[k] += h*p[k];
            __syncthreads();
            dPotential(k,nvar,b,n,q,dp);
            p[k] -= 0.5*h*dp[k];
            __syncthreads();
}

__device__ void energy(int nvar,float b, float n, float *q, float *p, float *e) {
            float x1,x2;
            int k = threadIdx.x;
            if (k==0) {x1=q[k]; x2=0.0;}
            else if (k==nvar-1) {x1=q[k]-q[k-1]; x2=-q[k];}
            else {x1=q[k]-q[k-1]; x2=0.0;}
            e[k] = 0.5*p[k]*p[k]+0.5*(x1*x1+x2*x2)+1.0/(n+1)*b*(powf(x1,n+1)+powf(x2,n+1));
            __syncthreads();
            int i = blockDim.x/2;
            while (i!=0) {
                if (k < i) e[k] += e[k+i];
                __syncthreads();
                i /= 2;
            }
}

extern __shared__ float shared[];

__global__ void verletMultiStepsKernel(int nsteps, float h, int nvar, float b, float n, float *q0, float *p0, float *en) {
            float *q = &shared[0];
            float *p = &q[nvar];
            float *dp = &p[nvar];
            float *e = &dp[nvar];
            int i;
            int k = threadIdx.x;
            q[k] = q0[k];
            p[k] = p0[k];
            __syncthreads();
            for (i=0; i<nsteps; i++) verletStep(k,h,nvar,b,n,q,p,dp);
            q0[k] = q[k];
            p0[k] = p[k];
            energy(nvar,b,n,q,p,e);
            if (k==0) en[0] = e[0];
}

__global__ void modeEnergyKernel(int nvar, float *q, float *p, float *A, float *omeg2, float *emode) {
            float *xi = &shared[0];
            float *dxi = &xi[nvar];
            int i = blockIdx.x; // mode
            int j = threadIdx.x; // variable
            int index = j+blockDim.x*i;
            xi[j] = A[index]*q[j];
            dxi[j] = A[index]*p[j];
            __syncthreads();
            int j1 = blockDim.x/2;
            while (j1!=0) {
                if (j < j1) {
                    xi[j] += xi[j+j1];
                    dxi[j] += dxi[j+j1];
                }
                __syncthreads();
                j1 /= 2;
            }
            if (j==0) emode[i] = 0.5*(dxi[0]*dxi[0]+omeg2[i]*xi[0]*xi[0]);
}

/*integrator step with global memory*/
__global__ void verletStep_1(float h, int nvar, float b, float n, float *q, float *p, float *dp) {
    int k = threadIdx.x+blockIdx.x*blockDim.x;
    if (k<nvar) {
        dPotential(k,nvar,b,n,q,dp);
        p[k] -= 0.5*h*dp[k];
        q[k] += h*p[k];
    }
}

__global__ void verletStep_2(float h, int nvar, float b, float n, float *q, float *p, float *dp) {
    int k = threadIdx.x+blockIdx.x*blockDim.x;
    if (k<nvar) {
        dPotential(k,nvar,b,n,q,dp);
        p[k] -= 0.5*h*dp[k];
    }
}

__global__ void energy_1(int nvar,float b, float n, float *q, float *p, float *en) {
    float x1,x2;
    float *e = &shared[0];
    int tid = threadIdx.x;
    e[tid] = 0.0;
    int k = threadIdx.x+blockIdx.x*blockDim.x;
    if (k<nvar) {
        if (k==0) {x1=q[k]; x2=0.0;}
        else if (k==nvar-1) {x1=q[k]-q[k-1]; x2=-q[k];}
        else {x1=q[k]-q[k-1]; x2=0.0;}
        e[tid] = 0.5*p[k]*p[k]+0.5*(x1*x1+x2*x2)+1.0/(n+1)*b*(powf(x1,n+1)+powf(x2,n+1));
        __syncthreads();
        int i = blockDim.x/2;
        while (i!=0) {
            if (tid < i) e[tid] += e[tid+i];
                __syncthreads();
            i /= 2;
        }
        en[blockIdx.x] = e[0];
    }
}
__global__ void energy_2(float *en) {
    int i = blockDim.x/2;
    int tid = threadIdx.x;
    while (i!=0) {
        if (tid < i) en[tid] += en[tid+i];
        __syncthreads();
        i /= 2;
     }
}

__global__ void modeEnergy_1(int nvar, float *q, float *p, float *A, float *xib, float *dxib) {
    float *xi = &shared[0];
    float *dxi = &xi[blockDim.x];
    int tid = threadIdx.x;
    int i = blockIdx.y; // mode
    int j = threadIdx.x+blockIdx.x*blockDim.x;
    xi[tid]= 0.0;
    dxi[tid] = 0.0;
    if (j>=nvar) return;
    int index = j+nvar*i;
    xi[tid] = A[index]*q[j];
    dxi[tid] = A[index]*p[j];
    __syncthreads();
    int j1 = blockDim.x/2;
            while (j1!=0) {
                if (tid < j1) {
                    xi[tid] += xi[tid+j1];
                    dxi[tid] += dxi[tid+j1];
                }
                __syncthreads();
                j1 /= 2;
    }
    if (tid==0) {
        xib[blockIdx.x+blockIdx.y*gridDim.x] = xi[0];
        dxib[blockIdx.x+blockIdx.y*gridDim.x] = dxi[0];
    }
    
}

__global__ void modeEnergy_2(float *xib, float *dxib, float *omeg2, float *emode) {
    int tid = threadIdx.x;
    int i = blockIdx.x; // mode
    int offset = i*blockDim.x;
    int j1 = blockDim.x/2;
            while (j1!=0) {
                if (tid < j1) {
                    xib[tid+offset] += xib[tid+offset+j1];
                    dxib[tid+offset] += dxib[tid+offset+j1];
                }
                __syncthreads();
                j1 /= 2;
    }
    if (tid==0) {
        emode[i] = 0.5*(dxib[offset]*dxib[offset]+omeg2[i]*xib[offset]*xib[offset]);
    }
}

extern "C" int fpu_open(int _nvar, int dev) {
    cudaDeviceProp prop;
    int count;
    int i,j;
    float r,s;
    float *A, *omeg2;
    nvar = _nvar;
    int device;
    HANDLE_ERROR(cudaGetDeviceCount(&count));
    if (count==0) return FPU_KERNEL_NO_CUDA;
    for (i=0; i<count; i++) {
        HANDLE_ERROR(cudaGetDeviceProperties(&prop,i));
        printf("Device %d : %s, compute capability : %d.%d\n",i,prop.name,prop.major,prop.minor); 
    }
    device = dev;
    printf("CUDA device selected : %d\n",device);
    HANDLE_ERROR(cudaSetDevice(device));
    maxThreads = prop.maxThreadsPerBlock;
    if (maxThreads >= nvar) {
        cpmode = CPMODE_LOCAL_MEM;
        nthreads = nvar;
        nblocks = 1;
    }
    else {
        cpmode = CPMODE_GLOBAL_MEM;
        nthreads = maxThreads;
        nblocks = (nvar+nthreads-1)/nthreads;
    }
    //if (prop.sharedMemPerBlock < nvar*4*sizeof(float)) return FPU_KERNEL_NVAR_TOO_LARGE;
    
    HANDLE_ERROR(cudaMalloc((void**)&dev_q,nvar*sizeof(float)));
    HANDLE_ERROR(cudaMalloc((void**)&dev_p,nvar*sizeof(float)));
    HANDLE_ERROR(cudaMalloc((void**)&dev_en,nblocks*sizeof(float)));
    HANDLE_ERROR(cudaMalloc((void**)&dev_omeg2,nvar*sizeof(float)));
    HANDLE_ERROR(cudaMalloc((void**)&dev_A,nvar*nvar*sizeof(float)));
    HANDLE_ERROR(cudaMalloc((void**)&dev_emode,nvar*sizeof(float)));
    if (cpmode==CPMODE_GLOBAL_MEM) {
        HANDLE_ERROR(cudaMalloc((void**)&dev_dp,nvar*sizeof(float)));
        HANDLE_ERROR(cudaMalloc((void**)&dev_xib,nvar*nblocks*sizeof(float)));
        HANDLE_ERROR(cudaMalloc((void**)&dev_dxib,nvar*nblocks*sizeof(float)));
    }
    A = (float*)malloc(nvar*nvar*sizeof(float));
    r = PI/(nvar+1);
    s = sqrt(2.0/(nvar+1));
    for (i=0; i<nvar; i++)
        for (j=0; j<nvar; j++) {
            A[i+nvar*j] = s*sin(r*(i+1)*(j+1));
        }
    HANDLE_ERROR(cudaMemcpy(dev_A,A,nvar*nvar*sizeof(float),cudaMemcpyHostToDevice));
    free(A);
    omeg2 = (float*)malloc(nvar*sizeof(float));
    r=PI/(2*(nvar+1));
    for (i=0; i<nvar; i++) omeg2[i] = 4*pow(sin((i+1)*r),2);
    HANDLE_ERROR(cudaMemcpy(dev_omeg2,omeg2,nvar*sizeof(float),cudaMemcpyHostToDevice));
    free(omeg2);
    return 0;
}

extern "C" int fpu_close() {
    HANDLE_ERROR(cudaFree(dev_q));
    HANDLE_ERROR(cudaFree(dev_p));
    HANDLE_ERROR(cudaFree(dev_en));
    HANDLE_ERROR(cudaFree(dev_omeg2));
    HANDLE_ERROR(cudaFree(dev_A));
    HANDLE_ERROR(cudaFree(dev_emode));
    if (cpmode==CPMODE_GLOBAL_MEM) {
        HANDLE_ERROR(cudaFree(dev_dp));
        HANDLE_ERROR(cudaFree(dev_xib));
        HANDLE_ERROR(cudaFree(dev_dxib));
    }
    return 0;
}


extern "C" int fpu_init(float *q0, float *p0) {
    HANDLE_ERROR(cudaMemcpy(dev_q,q0,nvar*sizeof(float),cudaMemcpyHostToDevice));
    HANDLE_ERROR(cudaMemcpy(dev_p,p0,nvar*sizeof(float),cudaMemcpyHostToDevice));
    return 0;
}

extern "C" int fpu_set_cst(float _b, float _n, float _h) {
    b = _b;
    n = _n;
    h = _h;
    return 0;
}

extern "C" int fpu_steps(int n1, float *q, float *p, float *e, float *emode, int nm) {
    int i1;
    dim3 blocks(nblocks,nm);
    if (cpmode==CPMODE_LOCAL_MEM) {
        verletMultiStepsKernel<<<nblocks,nthreads,nvar*4*sizeof(float)>>>(n1,h,nvar,b,n,dev_q,dev_p,dev_en);
        modeEnergyKernel<<<nm,nvar,nvar*2*sizeof(float)>>>(nvar,dev_q,dev_p,dev_A,dev_omeg2,dev_emode);
    }
    else {
        for (i1=0; i1<n1; i1++) {
            verletStep_1<<<nblocks,nthreads>>>(h,nvar,b,n,dev_q,dev_p,dev_dp);
            verletStep_2<<<nblocks,nthreads>>>(h,nvar,b,n,dev_q,dev_p,dev_dp);
            energy_1<<<nblocks,nthreads,nthreads*sizeof(float)>>>(nvar,b,n,dev_q,dev_p,dev_en);
            energy_2<<<1,nblocks>>>(dev_en);
            modeEnergy_1<<<blocks,nthreads,2*nthreads*sizeof(float)>>>(nvar,dev_q,dev_p,dev_A,dev_xib,dev_dxib);
            modeEnergy_2<<<nm,nblocks>>>(dev_xib,dev_dxib,dev_omeg2,dev_emode);
            
        }
    }
    HANDLE_ERROR(cudaMemcpy(q,dev_q,nvar*sizeof(float),cudaMemcpyDeviceToHost));
    HANDLE_ERROR(cudaMemcpy(p,dev_p,nvar*sizeof(float),cudaMemcpyDeviceToHost));
    HANDLE_ERROR(cudaMemcpy(e,dev_en,1*sizeof(float),cudaMemcpyDeviceToHost));
    HANDLE_ERROR(cudaMemcpy(emode,dev_emode,nvar*sizeof(float),cudaMemcpyDeviceToHost));
    HANDLE_ERROR(cudaThreadSynchronize());
    return 0;
}

/*FPU chain with periodic boundary conditions*/
__device__ void dPotential_per(int k, int nvar, float b, float n, float *q, float *dp) {
            float x1,x2;
            if (k==0) {x1=q[0]-q[nvar-1]; x2=q[1]-q[0];}
            else if (k==nvar-1) {x1 = q[k]-q[k-1]; x2=-q[k]+q[0];}
            else {x1=q[k]-q[k-1]; x2=-q[k]+q[k+1];}
            dp[k] = x1-x2+b*(powf(x1,n)-powf(x2,n));
}
__device__ void verletStep_per(int k, float h, int nvar, float b, float n, float *q, float *p, float *dp) {
            dPotential_per(k,nvar,b,n,q,dp);
            p[k] -= 0.5*h*dp[k];
            q[k] += h*p[k];
            __syncthreads();
            dPotential_per(k,nvar,b,n,q,dp);
            p[k] -= 0.5*h*dp[k];
            __syncthreads();
}

__device__ void energy_per(int nvar,float b, float n, float *q, float *p, float *e) {
            float x1;
            int k = threadIdx.x;
            if (k==0) x1=q[k]-q[nvar-1];
            else x1=q[k]-q[k-1];
            e[k] = 0.5*p[k]*p[k]+0.5*(x1*x1)+1.0/(n+1)*b*powf(x1,n+1);
            __syncthreads();
            int i = blockDim.x/2;
            while (i!=0) {
                if (k < i) e[k] += e[k+i];
                __syncthreads();
                i /= 2;
            }
}

__global__ void verletMultiStepsKernel_per(int nsteps, float h, int nvar, float b, float n, float *q0, float *p0, float *en) {
            float *q = &shared[0];
            float *p = &q[nvar];
            float *dp = &p[nvar];
            float *e = &dp[nvar];
            int i;
            int k = threadIdx.x;
            q[k] = q0[k];
            p[k] = p0[k];
            __syncthreads();
            for (i=0; i<nsteps; i++) verletStep_per(k,h,nvar,b,n,q,p,dp);
            q0[k] = q[k];
            p0[k] = p[k];
            energy_per(nvar,b,n,q,p,e);
            if (k==0) en[0] = e[0];
}
__global__ void modeEnergyKernel_per(int nvar, float *q, float *p, float *reA, float *imA, float *omeg2, float *emode) {
            float *rexi = &shared[0];
            float *imxi = &rexi[nvar];
            float *redxi = &imxi[nvar];
            float *imdxi = &redxi[nvar];
            int i = blockIdx.x; // mode
            int j = threadIdx.x; // variable
            int index = j+blockDim.x*i;
            rexi[j] = reA[index]*q[j];
            imxi[j] = imA[index]*q[j];
            redxi[j] = reA[index]*p[j];
            imdxi[j] = imA[index]*p[j];
            __syncthreads();
            int j1 = blockDim.x/2;
            while (j1!=0) {
                if (j < j1) {
                    rexi[j] += rexi[j+j1];
                    imxi[j] += imxi[j+j1];
                    redxi[j] += redxi[j+j1];
                    imdxi[j] += imdxi[j+j1];
                }
                __syncthreads();
                j1 /= 2;
            }
            if (j==0) emode[i] = 0.5*(redxi[0]*redxi[0]+imdxi[0]*imdxi[0]+omeg2[i]*(rexi[0]*rexi[0]+imxi[0]*imxi[0]));
}

/*integrator step with global memory*/
__global__ void verletStep_per_1(float h, int nvar, float b, float n, float *q, float *p, float *dp) {
    int k = threadIdx.x+blockIdx.x*blockDim.x;
    if (k<nvar) {
        dPotential_per(k,nvar,b,n,q,dp);
        p[k] -= 0.5*h*dp[k];
        q[k] += h*p[k];
    }
}

__global__ void verletStep_per_2(float h, int nvar, float b, float n, float *q, float *p, float *dp) {
    int k = threadIdx.x+blockIdx.x*blockDim.x;
    if (k<nvar) {
        dPotential_per(k,nvar,b,n,q,dp);
        p[k] -= 0.5*h*dp[k];
    }
}

__global__ void energy_per_1(int nvar,float b, float n, float *q, float *p, float *en) {
    float x1;
    float *e = &shared[0];
    int tid = threadIdx.x;
    e[tid] = 0.0;
    int k = threadIdx.x+blockIdx.x*blockDim.x;
    if (k<nvar) {
        if (k==0) x1=q[k]-q[nvar-1];
        else x1=q[k]-q[k-1];
        e[tid] = 0.5*p[k]*p[k]+0.5*x1*x1+1.0/(n+1)*b*powf(x1,n+1);
        __syncthreads();
        int i = blockDim.x/2;
        while (i!=0) {
            if (tid < i) e[tid] += e[tid+i];
                __syncthreads();
            i /= 2;
        }
        en[blockIdx.x] = e[0];
    }
}
__global__ void energy_per_2(float *en) {
    int i = blockDim.x/2;
    int tid = threadIdx.x;
    while (i!=0) {
        if (tid < i) en[tid] += en[tid+i];
        __syncthreads();
        i /= 2;
     }
}

__global__ void modeEnergy_per_1(int nvar, float *q, float *p, float *reA, float *imA, float *rexib, float *imxib, float *redxib, float *imdxib) {
    float *rexi = &shared[0];
    float *imxi = &rexi[blockDim.x];
    float *redxi = &imxi[blockDim.x];
    float *imdxi = &redxi[blockDim.x];
    int tid = threadIdx.x;
    int i = blockIdx.y; // mode
    int j = threadIdx.x+blockIdx.x*blockDim.x;
    rexi[tid]= 0.0;
    imxi[tid] = 0.0;
    redxi[tid] = 0.0;
    imdxi[tid] = 0.0;
    if (j>=nvar) return;
    int index = j+nvar*i;
    rexi[tid] = reA[index]*q[j];
    imxi[tid] = imA[index]*q[j];
    redxi[tid] = reA[index]*p[j];
    imdxi[tid] = imA[index]*p[j];
    __syncthreads();
    int j1 = blockDim.x/2;
            while (j1!=0) {
                if (tid < j1) {
                    rexi[tid] += rexi[tid+j1];
                    imxi[tid] += imxi[tid+j1];
                    redxi[tid] += redxi[tid+j1];
                    imdxi[tid] += imdxi[tid+j1];
                }
                __syncthreads();
                j1 /= 2;
    }
    if (tid==0) {
        rexib[blockIdx.x+blockIdx.y*gridDim.x] = rexi[0];
        imxib[blockIdx.x+blockIdx.y*gridDim.x] = imxi[0];
        redxib[blockIdx.x+blockIdx.y*gridDim.x] = redxi[0];
        imdxib[blockIdx.x+blockIdx.y*gridDim.x] = imdxi[0];
    }
    
}

__global__ void modeEnergy_per_2(float *rexib, float *imxib, float *redxib, float *imdxib, float *omeg2, float *emode) {
    int tid = threadIdx.x;
    int i = blockIdx.x; // mode
    int offset = i*blockDim.x;
    int j1 = blockDim.x/2;
            while (j1!=0) {
                if (tid < j1) {
                    rexib[tid+offset] += rexib[tid+offset+j1];
                    imxib[tid+offset] += imxib[tid+offset+j1];
                    redxib[tid+offset] += redxib[tid+offset+j1];
                    imdxib[tid+offset] += imdxib[tid+offset+j1];
                }
                __syncthreads();
                j1 /= 2;
    }
    if (tid==0) {
        emode[i] = 0.5*(redxib[offset]*redxib[offset]+imdxib[offset]*imdxib[offset]+omeg2[i]*(rexib[offset]*rexib[offset]+imxib[offset]*imxib[offset]));
    }
}


extern "C" int fpuper_open(int _nvar, int dev) {
    cudaDeviceProp prop;
    int count;
    int i,j;
    float r,s;
    float *reA, *imA, *omeg2;
    nvar = _nvar;
    int device;
    HANDLE_ERROR(cudaGetDeviceCount(&count));
    if (count==0) return FPU_KERNEL_NO_CUDA;
    for (i=0; i<count; i++) {
        HANDLE_ERROR(cudaGetDeviceProperties(&prop,i));
        printf("Device %d : %s, compute capability : %d.%d\n",i,prop.name,prop.major,prop.minor); 
    }
    device = dev;
    printf("CUDA device selected : %d\n",device);
    HANDLE_ERROR(cudaSetDevice(device));
    HANDLE_ERROR(cudaSetDevice(device));
    maxThreads = prop.maxThreadsPerBlock;
    if (maxThreads >= nvar) {
        cpmode = CPMODE_LOCAL_MEM;
        nthreads = nvar;
        nblocks = 1;
    }
    else {
        cpmode = CPMODE_GLOBAL_MEM;
        nthreads = maxThreads;
        nblocks = (nvar+nthreads-1)/nthreads;
    }
    //if (prop.sharedMemPerBlock < nvar*4*sizeof(float)) return FPU_KERNEL_NVAR_TOO_LARGE;
    HANDLE_ERROR(cudaMalloc((void**)&dev_q,nvar*sizeof(float)));
    HANDLE_ERROR(cudaMalloc((void**)&dev_p,nvar*sizeof(float)));
    HANDLE_ERROR(cudaMalloc((void**)&dev_en,1*sizeof(float)));
    HANDLE_ERROR(cudaMalloc((void**)&dev_omeg2,nvar*sizeof(float)));
    HANDLE_ERROR(cudaMalloc((void**)&dev_reA,nvar*nvar*sizeof(float)));
    HANDLE_ERROR(cudaMalloc((void**)&dev_imA,nvar*nvar*sizeof(float)));
    HANDLE_ERROR(cudaMalloc((void**)&dev_emode,nvar*sizeof(float)));
    if (cpmode==CPMODE_GLOBAL_MEM) {
        HANDLE_ERROR(cudaMalloc((void**)&dev_dp,nvar*sizeof(float)));
        HANDLE_ERROR(cudaMalloc((void**)&dev_rexib,nvar*nblocks*sizeof(float)));
        HANDLE_ERROR(cudaMalloc((void**)&dev_redxib,nvar*nblocks*sizeof(float)));
        HANDLE_ERROR(cudaMalloc((void**)&dev_imxib,nvar*nblocks*sizeof(float)));
        HANDLE_ERROR(cudaMalloc((void**)&dev_imdxib,nvar*nblocks*sizeof(float)));
    }
    reA = (float*)malloc(nvar*nvar*sizeof(float));
    imA = (float*)malloc(nvar*nvar*sizeof(float));
    r = PI*2/nvar;
    s = sqrt(1.0/nvar);
    for (i=0; i<nvar; i++)
        for (j=0; j<nvar; j++) {
            reA[i+nvar*j] = s*cos(r*(i+1)*(j+1));
            imA[i+nvar*j] = -s*sin(r*(i+1)*(j+1));
        }
    HANDLE_ERROR(cudaMemcpy(dev_reA,reA,nvar*nvar*sizeof(float),cudaMemcpyHostToDevice));
    HANDLE_ERROR(cudaMemcpy(dev_imA,imA,nvar*nvar*sizeof(float),cudaMemcpyHostToDevice));
    free(reA); free(imA);
    omeg2 = (float*)malloc(nvar*sizeof(float));
    r=PI/nvar;
    for (i=0; i<nvar; i++) omeg2[i] = 4*pow(sin((i+1)*r),2);
    HANDLE_ERROR(cudaMemcpy(dev_omeg2,omeg2,nvar*sizeof(float),cudaMemcpyHostToDevice));
    free(omeg2);
    return 0;
}

extern "C" int fpuper_close() {
    HANDLE_ERROR(cudaFree(dev_q));
    HANDLE_ERROR(cudaFree(dev_p));
    HANDLE_ERROR(cudaFree(dev_en));
    HANDLE_ERROR(cudaFree(dev_omeg2));
    HANDLE_ERROR(cudaFree(dev_reA));
    HANDLE_ERROR(cudaFree(dev_imA));
    HANDLE_ERROR(cudaFree(dev_emode));
    if (cpmode==CPMODE_GLOBAL_MEM) {
        HANDLE_ERROR(cudaFree(dev_dp));
        HANDLE_ERROR(cudaFree(dev_rexib));
        HANDLE_ERROR(cudaFree(dev_redxib));
        HANDLE_ERROR(cudaFree(dev_imxib));
        HANDLE_ERROR(cudaFree(dev_imdxib));
    }
    return 0;
}

extern "C" int fpuper_init(float *q0, float *p0) {
    HANDLE_ERROR(cudaMemcpy(dev_q,q0,nvar*sizeof(float),cudaMemcpyHostToDevice));
    HANDLE_ERROR(cudaMemcpy(dev_p,p0,nvar*sizeof(float),cudaMemcpyHostToDevice));
    return 0;
}

extern "C" int fpuper_set_cst(float _b, float _n, float _h) {
    b = _b;
    n = _n;
    h = _h;
    return 0;
}

extern "C" int fpuper_steps(int n1, float *q, float *p, float *e, float *emode, int nm) {
    int i1;
    dim3 blocks(nblocks,nm);
    if (cpmode==CPMODE_LOCAL_MEM) {
        verletMultiStepsKernel_per<<<1,nvar,nvar*4*sizeof(float)>>>(n1,h,nvar,b,n,dev_q,dev_p,dev_en);
        modeEnergyKernel_per<<<nm,nvar,nvar*4*sizeof(float)>>>(nvar,dev_q,dev_p,dev_reA,dev_imA,dev_omeg2,dev_emode);
    }
    else {
        for (i1=0; i1<n1; i1++) {
            verletStep_per_1<<<nblocks,nthreads>>>(h,nvar,b,n,dev_q,dev_p,dev_dp);
            verletStep_per_2<<<nblocks,nthreads>>>(h,nvar,b,n,dev_q,dev_p,dev_dp);
            energy_per_1<<<nblocks,nthreads,nthreads*sizeof(float)>>>(nvar,b,n,dev_q,dev_p,dev_en);
            energy_per_2<<<1,nblocks>>>(dev_en);
            modeEnergy_per_1<<<blocks,nthreads,4*nthreads*sizeof(float)>>>(nvar,dev_q,dev_p,dev_reA,dev_imA, dev_rexib, dev_imxib, dev_redxib, dev_imdxib);
            modeEnergy_per_2<<<nm,nblocks>>>(dev_rexib, dev_imxib, dev_redxib, dev_imdxib, dev_omeg2,dev_emode);
        }
    
    }
    HANDLE_ERROR(cudaMemcpy(q,dev_q,nvar*sizeof(float),cudaMemcpyDeviceToHost));
    HANDLE_ERROR(cudaMemcpy(p,dev_p,nvar*sizeof(float),cudaMemcpyDeviceToHost));
    HANDLE_ERROR(cudaMemcpy(e,dev_en,1*sizeof(float),cudaMemcpyDeviceToHost));
    HANDLE_ERROR(cudaMemcpy(emode,dev_emode,nvar*sizeof(float),cudaMemcpyDeviceToHost));
    HANDLE_ERROR(cudaThreadSynchronize());
    return 0;
}
